package be.cylab.java.wowa.training;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

/**
 * Abstract class Solution.
 */
public abstract class AbstractSolution
        implements Comparable<AbstractSolution>, Cloneable {

    protected List<Double> weights_w;
    protected List<Double> weights_p;
    protected double fitness_score = Double.POSITIVE_INFINITY;

    /**
     * SolutionDistance constructor. Needs weight number as parameter
     *
     * @param weights_number
     */
    public AbstractSolution(final int weights_number) {
        this.weights_w = new ArrayList<>();
        this.weights_p = new ArrayList<>();
        for (int i = 0; i < weights_number; i++) {
            this.weights_w.add(Math.random());
            this.weights_p.add(Math.random());
        }
        this.normalize();
    }

    /**
     * Constructor solution distance. Made for test to initialize the Random
     * Generator
     *
     * @param weights_number int
     * @param seed           int
     */

    public AbstractSolution(final int weights_number, final int seed) {
        Random rnd = new Random(seed);
        this.weights_w = new ArrayList<>();
        this.weights_p = new ArrayList<>();
        for (int i = 0; i < weights_number; i++) {
            this.weights_w.add(rnd.nextDouble());
            this.weights_p.add(rnd.nextDouble());
        }
        this.normalize();
    }

    abstract void computeScoreTo(List<double[]> data, double[] expected);

    /**
     * @return
     */
    @Override
    public final String toString() {
        return "SolutionDistance{"
                + "weights_w=" + Arrays.toString(weights_w.toArray())
                + ", weights_p=" + Arrays.toString(weights_p.toArray())
                + ", fitness score=" + Math.abs(fitness_score)
                + '}';
    }

    /**
     * Change a random gene in a weight vector.
     * The gene is changed is the probability is higher than a random double
     *
     * @param probability
     */
    final void randomlyMutateWithProbability(final double probability) {
        double tos = Math.random();
        if (tos > probability) {
            // do nothing
            return;
        }

        //new weight value and position
        double new_weight_value = Math.random();
        int random_index = Utils.randomInteger(0, this.weights_p.size() - 1);
        //Select w or p weights
        int weight_selection = Utils.randomInteger(0, 1);
        if (weight_selection == 0) {
            this.weights_w.set(random_index, new_weight_value);
        } else {
            this.weights_p.set(random_index, new_weight_value);
        }
        this.normalize();

    }

    /**
     * @param solution
     * @return int
     */
    @Override
    public final int compareTo(final AbstractSolution solution) {
        if (Double.compare(this.getFitnessScore(),
                solution.getFitnessScore()) > 0) {
            return 1;
        } else if (Double.compare(this.getFitnessScore(),
                solution.getFitnessScore()) < 0) {
            return -1;
        } else {
            return 0;
        }
    }

    /**
     * Function to normalize SolutionDistance weights.
     * Weights must be between 0 and 1 and the sum of the weight in a vector
     * must be equal to 1
     */
    final void normalize() {
        this.weights_w = Utils.normalizeWeights(this.weights_w);
        this.weights_p = Utils.normalizeWeights(this.weights_p);
    }

    /**
     * @return
     */
    public final double getFitnessScore() {
        return fitness_score;
    }


    /**
     * @return
     */
    public final List<Double> getWeightsW() {
        return this.weights_w;
    }

    /**
     * @return
     */
    public final List<Double> getWeightsP() {
        return this.weights_p;
    }

    /**
     * @param weights_w
     */
    final void setWeightsW(final List<Double> weights_w) {
        this.weights_w = weights_w;
    }

    /**
     * @param weights_p
     */
    final void setWeightsP(final List<Double> weights_p) {
        this.weights_p = weights_p;
    }


    /**
     * Copy.
     *
     * @return
     * @throws java.lang.CloneNotSupportedException if clone is not supported
     */
    public final SolutionDistance clone() throws CloneNotSupportedException {
        return (SolutionDistance) super.clone();
    }

}
